"""
Main for nbsplit

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

import copy
from argparse import _ArgumentGroup, SUPPRESS
from pathlib import Path
from typing import Optional, List

from nbformat import NotebookNode

from nbtoolbelt.arguments import split_csv, join_csv
from nbtoolbelt.counting import REQUIRED_CELL_FIELDS
from nbtoolbelt.printing import print_dict
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.toolbaseapp import Tool

TOOL = 'split'


class SplitTool(Tool):
    """Split each notebook.
    """
    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            action=TOOL,
            description="""Split Jupyter notebooks into code cells and other cells."""
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Split notebook nb into markdown, code, and raw cells.

        .. note::

          **Modifies**: ``nb``

          **Side effect**: writes new notebook files

        :param nb: notebook to split
        :param nb_path: path to ``nb``
        :return: sequence of resulting (notebook, notebook-path) pairs
        """
        args = self._args

        nb_result = {}
        freq = {'total': len(nb.cells)}
        nb_result_path = {}

        # split the notebook into a part for each cell type
        for cell_type in args.split_cell_types:
            nb_result[cell_type] = copy.deepcopy(nb)  # notebook for cell type
            nb_result[cell_type].cells = [cell for cell in nb.cells if cell.cell_type == cell_type]
            freq[cell_type] = len(nb_result[cell_type].cells)

        # print statistics
        if not args.quiet:
            print_dict(freq, 'Split cell statistics')
        self._aggregate['outputs'][-1].update(freq)

        # set up destination paths
        for cell_type in args.split_cell_types:
            name = nb_path.stem + getattr(args, 'split_' + cell_type + '_result_name') + nb_path.suffix
            nb_result_path[cell_type] = nb_path.with_name(name)
            if args.debug and freq[cell_type]:
                print('Destination of {} cells:'.format(cell_type), nb_result_path[cell_type])

        # only write parts that have cells
        result = [(nb_result[cell_type], nb_result_path[cell_type])
                  for cell_type in args.split_cell_types
                  if freq[cell_type]]

        return result

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-t', '--split-cell-types', default=SUPPRESS, type=split_csv,
                           metavar='TYPES',
                           help='comma-separated list of cell types to split'
                                " (default: '{}')".format(join_csv(self._args.split_cell_types)))

    def check_and_adjust_arguments(self) -> None:
        bad = set(self._args.split_cell_types).difference(REQUIRED_CELL_FIELDS.keys())
        if bad:
            raise ValueError('Invalid cell types with -t: {}'.format(join_csv(bad)))

    def print_tool_args(self) -> None:
        print('  Split cell types: {}'.format(join_csv(self._args.split_cell_types)))


def main(cli_args: Optional[List[str]] = None) -> int:
    return SplitTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Functions for rendering

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.

Parts of this code are reworked from the rendernb checklet for Momotor <momotor.org>.
"""

from argparse import Namespace
from typing import Any, Dict

from nbconvert import HTMLExporter
from nbformat import NotebookNode
from traitlets.config import Config

from .inline_attachments import InlineAttachmentsPreprocessor


# ADDITIONAL_STYLES = dedent("""\
#     div.inner_cell, div.output_subarea {
#       flex: 1 auto !important;
#     }
#     """)
#
# HTML_FRAME = dedent("""\
#     <html>
#     <head>
#     <style>
#     {css}
#     </style>
#     <script>
#     {javascript}
#     </script>
#     </head>
#     <body>
#     {body}
#     </body>
#     </html>
#     """)


def render_nb(notebook: NotebookNode, args: Namespace) -> Dict[str, Any]:
    """Render notebook as html.
    Uses ``args.template`` as template name.

    :param notebook: notebook to render
    :param args: options
    :return: html
    """
    resources = {}

    iapp = InlineAttachmentsPreprocessor()
    notebook, resources = iapp.preprocess(notebook, resources)

    c = Config()
    c.HTMLExporter.template_name = args.template
    # TODO also set extra_template_basedirs from option?
    html_exporter = HTMLExporter(config=c)

    html, resources = html_exporter.from_notebook_node(notebook)

    # properties = {
    #     'body': html,
    #     'css': '',
    #     'javascript': ''
    # }

    # return HTML_FRAME.format(**properties)
    return html



"""
Main for nbclean

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from argparse import _ArgumentGroup, SUPPRESS
from pathlib import Path
from typing import Optional, List, Set

from nbformat import NotebookNode

from nbtoolbelt.arguments import split_csv, NegatableAction, join_csv
from nbtoolbelt.cleaning import clean_nb
from nbtoolbelt.counting import REQUIRED_NB_METADATA_FIELDS
from nbtoolbelt.printing import print_dict
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.toolbaseapp import Tool

TOOL = 'clean'


class CleanTool(Tool):
    """Clean each notebook.
    """
    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            action=TOOL,
            description="""Clean Jupyter notebooks by deleting or resetting selected elements."""
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Clean notebook nb.

        .. note:: **Modifies**: ``nb``

        :param nb: notebook to clean
        :param nb_path: path to ``nb``
        :return: sequence of resulting (notebook, notebook-path) pairs
        """
        args = self._args

        # do the cleaning, and collect some statistics
        freq = clean_nb(nb, args)

        # show statistics
        if not args.quiet:
            print_dict(freq, 'Cleaning counts')
        self._aggregate['outputs'][-1].update(freq)

        # set up destination path
        if args.inplace:
            nb_clean_path = nb_path
        else:
            nb_clean_path = nb_path.with_name(nb_path.stem + args.clean_result_name + nb_path.suffix)

        if args.debug:
            print('Clean destination:', nb_clean_path)

        return [(nb, nb_clean_path)]

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-g', '--global',
                           dest='clean_notebook_metadata_fields', default=SUPPRESS, type=split_csv,
                           metavar='FIELDS',
                           help='comma-separated list of fields to remove from metadata of notebook' +
                                " (default: '{}')".format(join_csv(self._args.clean_notebook_metadata_fields)))
        group.add_argument('-m', '--metadata',
                           dest='clean_cell_metadata_fields', default=SUPPRESS, type=split_csv,
                           metavar='FIELDS',
                           help='comma-separated list of fields to remove from metadata of all cells' +
                                " (default: '{}')".format(join_csv(self._args.clean_cell_metadata_fields)))
        group.add_argument('-t', '--tags', dest='clean_tags', default=SUPPRESS, type=split_csv,
                           metavar='TAGS',
                           help='comma-separated list of tags to remove from all cells;'
                                " use '-m tags' to remove all tags" +
                                " (default: '{}')".format(join_csv(self._args.clean_tags)))
        group.add_argument('-e', '--empty-cells',
                           action=NegatableAction, dest='clean_empty_cells', default=SUPPRESS,
                           help='delete cells with empty source, i.e. with whitespace only' +
                                ' (default: {})'.format(self._args.clean_empty_cells))
        group.add_argument('-o', '--outputs',
                           action=NegatableAction, dest='clean_outputs', default=SUPPRESS,
                           help='clean all outputs from all code cells ' +
                                ' (default: {})'.format(self._args.clean_outputs))

    def check_and_adjust_arguments(self):
        args = self._args

        nmf = set(args.clean_notebook_metadata_fields)  # type: Set[str]
        if nmf:
            intersection = nmf.intersection(REQUIRED_NB_METADATA_FIELDS)
            if intersection:
                print('WARNING: {} are required notebook metadata fields'.format(intersection))
                print('  These will not be cleaned')
                nmf.difference_update(REQUIRED_NB_METADATA_FIELDS)

    def print_tool_args(self) -> None:
        args = self._args

        if args.clean_notebook_metadata_fields:
            print("  Deleting notebook metadata fields: {}".format(args.clean_notebook_metadata_fields))
        if args.clean_cell_metadata_fields:
            print("  Deleting cell metadata fields: {}".format(args.clean_cell_metadata_fields))
        if args.clean_tags:
            print("  Deleting tags: {}".format(args.clean_tags))
        if args.clean_empty_cells:
            print('  Deleting cells with empty source')
        if args.clean_outputs:
            print('  Deleting outputs from all code cells')


def main(cli_args: Optional[List[str]] = None) -> int:
    return CleanTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Tool nbdump; dump cell sources of notebooks

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from argparse import _ArgumentGroup, SUPPRESS
from pathlib import Path
from typing import Optional, List

from nbformat import NotebookNode

from nbtoolbelt.arguments import NegatableAction, split_csv, join_csv
from nbtoolbelt.counting import REQUIRED_CELL_FIELDS, nb_other_metadata
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.toolbaseapp import Tool

TOOL = 'dump'


class DumpTool(Tool):
    """Show all source lines of each notebook.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            description='Dump cell sources of Jupyter notebooks.',
            action=TOOL
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Dump cell sources of notebook.

        :param nb: notebook to dump
        :param nb_path: notebook's path
        :return: empty result
        """
        info_prefix = self._args.dump_info_prefix
        line_indent = ' ' * self._args.dump_source_line_indent
        cell_spacing = '\n' * self._args.dump_cell_spacing

        if self._args.dump_notebook_info:
            print('{} nbformat: {:1d}.{:1d}'
                  .format(info_prefix, nb.get('nbformat', 1), nb.get('nbformat_minor', 0)), end='')
            if 'kernelspec' in nb.metadata:
                print(' | kernel: {}'.format(nb.metadata.kernelspec.name), end='')
            if 'language_info' in nb.metadata:
                l_i = nb.metadata.language_info
                print(' | language: {} {}'.format(l_i.name, l_i.version), end='')
            print(' {}'.format(info_prefix))
            print('{} cells: {}'.format(info_prefix, len(nb.cells)), end='')
            other_metadata = nb_other_metadata(nb)
            if other_metadata:
                print(' | metadata: {}'.format(join_csv(other_metadata.keys())), end='')
            print(' {}'.format(info_prefix))

        spacing = ''

        for index, cell in enumerate(nb.cells):
            ct = cell.cell_type
            if ct in self._args.dump_cell_types:
                print(spacing, end='')
                spacing = cell_spacing
                if self._args.dump_cell_info:
                    print('{} cell {}: {}'.format(info_prefix, index, ct), end='')
                    metadata = set(cell.metadata.keys())
                    metadata.discard('tags')
                    if metadata:
                        print(' | metadata: {}'.format(','.join(metadata)), end='')
                    if 'tags' in cell.metadata:
                        tags = ','.join(cell.metadata.tags)
                        print(' | tags: {}'.format(tags), end='')
                    if (ct == 'markdown' or ct == 'raw') and 'attachments' in cell:
                        print(' | attachments: {}'.format(len(cell.attachments)), end='')
                    if ct == 'code' and cell.outputs:
                        print(' | outputs: {}'.format(len(cell.outputs)), end='')
                    print(' {}'.format(info_prefix))
                if self._args.dump_sources and cell.source:
                    for line_index, line in enumerate(cell.source.split('\n')):
                        if self._args.dump_source_line_numbers:
                            print('{}{:>2}| {}'.format(line_indent, line_index + 1, line), sep='')
                        else:
                            print('{}{}'.format(line_indent, line), sep='')

        return []  # no notebooks produced

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-g', '--notebook-info',
                           action=NegatableAction, default=SUPPRESS, dest='dump_notebook_info',
                           help='show global notebook information' +
                                ' (default: {})'.format(self._args.dump_notebook_info))
        group.add_argument('-c', '--cell-info',
                           action=NegatableAction, default=SUPPRESS, dest='dump_cell_info',
                           help='show cell information' +
                                ' (default: {})'.format(self._args.dump_cell_info))
        group.add_argument('-s', '--sources',
                           action=NegatableAction, default=SUPPRESS, dest='dump_sources',
                           help='show cell sources' +
                           ' (default: {})'.format(self._args.dump_sources))
        group.add_argument('-t', '--cell-types', default=SUPPRESS, type=split_csv, dest='dump_cell_types',
                           metavar='TYPES',
                           help='comma-separated list of cell types to dump' +
                                " (default: '{}')".format(join_csv(self._args.dump_cell_types)))
        group.add_argument('-p', '--prefix',
                           default=SUPPRESS, dest='dump_info_prefix', metavar='STRING',
                           help='pre/postfix for information lines' +
                                " (default: '{}')".format(self._args.dump_info_prefix))
        group.add_argument('-i', '--indent',
                           default=SUPPRESS, type=int, dest='dump_source_line_indent', metavar='INT',
                           help='indentation level for source lines' +
                                ' (default: {})'.format(self._args.dump_source_line_indent))
        group.add_argument('-l', '--line-numbers',
                           action=NegatableAction, default=SUPPRESS, dest='dump_source_line_numbers',
                           help='show source line numbers' +
                                ' (default: {})'.format(self._args.dump_source_line_numbers))
        group.add_argument('-e', '--cell-spacing',
                           default=SUPPRESS, type=int, dest='dump_cell_spacing', metavar='INT',
                           help='number of empty lines between cells' +
                                ' (default: {})'.format(self._args.dump_cell_spacing))

    def check_and_adjust_arguments(self) -> None:
        bad = set(self._args.dump_cell_types).difference(REQUIRED_CELL_FIELDS.keys())
        if bad:
            raise ValueError('Invalid cell types with -t: {}'.format(join_csv(bad)))

    def print_tool_args(self) -> None:
        print('  Dump notebook info: {}'.format(self._args.dump_notebook_info))
        print('  Dump cell info: {}'.format(self._args.dump_cell_info))
        print('  Dump cell sources: {}'.format(self._args.dump_sources))
        print('  Dump cell types: {}'.format(join_csv(self._args.dump_cell_types)))
        print('  Prefix for information lines: {}'.format(self._args.dump_info_prefix))
        print('  Source line indentation level: {}'.format(self._args.dump_source_line_indent))
        print('  Show source line numbers: {}'.format(self._args.dump_source_line_numbers))
        print('  Cell spacing lines: {}'.format(self._args.dump_cell_spacing))


def main(cli_args: Optional[List[str]] = None) -> int:
    return DumpTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Tool nbvalidate

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

import sys
from argparse import _ArgumentGroup
from pathlib import Path
from typing import Optional, List

from nbformat import NotebookNode

from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.toolbaseapp import Tool
from nbtoolbelt.validating import validate_nb

TOOL = 'validate'


class ValidateTool(Tool):
    """Validate each notebook, and report the validation results.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            action=TOOL,
            description="""Validate Jupyter notebooks."""
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Validate notebook nb.

        :param nb: notebook to validate
        :param nb_path: path to ``nb``
        :return: empty result
        """
        valid = validate_nb(nb, getattr(self._args, 'assert'))

        if not valid or self._args.verbose:
            print('{} notebook structure is {}'.format(nb_path.name, 'valid' if valid else 'INVALID'))
        self._aggregate['outputs'][-1]['valid'] = valid

        return []

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        pass

    def print_tool_args(self) -> None:
        print('  Assert mode: {}'.format(getattr(self._args, 'assert')))


def main(cli_args: Optional[List[str]] = None) -> int:
    return ValidateTool().main(cli_args)


if __name__ == "__main__":
    sys.exit(main())



"""
Functions for punching

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

import copy
import re
from argparse import Namespace
from collections import defaultdict
from typing import Any, Tuple, List, Dict, Set, Optional

from nbformat import NotebookNode

from .notebook_io import cell_lines

TEST = False


def punch_via_tags(nb: NotebookNode, args: Namespace) -> Tuple[Dict[str, int], NotebookNode, NotebookNode]:
    """Punch notebook via tags approach.

    .. note:: **Modifies**: nb

    :param nb: notebook to punch
    :param args: options
    :return: punch counts, punched notebook, chads notebook
    """
    # initialize punch counts
    freq = defaultdict(int)

    # create new notebook for the chads (content of cleared cells)
    nb_chads = copy.deepcopy(nb)
    nb_chads.cells = []

    # traverse all cells, and clear source if it has a tag in the trigger set
    # first, add to-be-cleared cell to chads notebook
    for cell in nb.cells:
        ct = cell.cell_type
        freq[ct] += 1
        if 'tags' in cell.metadata:
            if set(cell.metadata.tags).intersection(set(args.tags)):  # trigger tag present
                nb_chads.cells.append(copy.deepcopy(cell))
                cell.source = args.filling[ct] if args.fill else []
                if 'outputs' in cell:
                    cell.outputs = []
                if 'execution_count' in cell:
                    cell.execution_count = None
                freq[ct + ' holes'] += 1

    return dict(freq), nb, nb_chads


class PunchBaseProcessor:
    """Base class for ``punch`` processors,
    using the Template Method pattern.
    The base class takes care of parsing the marking structure.
    The concrete subclasses provide semantics for marking structure.

    Template methods:

    * ``punch_nb()``: main entry point; calls ``punch_line()`` and hook methods
      * ``pre_cell()``,
      * ``post_cell()``, and
      * ``post_nb()``

    * ``punch_line()``: calls hook methods
      * ``punch_begin_marker_line()``,
      * ``punch_end_marker_line()``,
      * ``punch_inside_line()``, and
      * ``punch_outside_line()``

    Concrete processors override the following hook methods:

    * ``__init__()`` to initialize; must call ``super()`` at begin
    * ``pre_cell()``
    * ``punch_begin_marker_line()``
    * ``punch_end_marker_line()``
    * ``punch_inside_line()``
    * ``punch_outside_line()``
    * ``post_cell()``
    * ``post_nb()``

    A concrete class is typically instantiated and used to process a notebook in one go:

    * ``ConcretePunchProcessor(args, nb, name).punch_nb()``
    """

    # TODO consider whether nb should be param of punch_nb(); this will require pre_nb() hook
    # TODO consider whether punch_nb() should be named __call__()

    def __init__(self, nb: NotebookNode, args: Namespace, name: str):
        """Initialize punch processor for given notebook and options.

        Some PunchProcessors may modify args, e.g. add ``source_chads``.

        :param nb: notebook to punch
        :param args: options
        :param name: notebook name used in messages
        """
        self.args = args
        self.nb = nb
        self.name = name

        # initialize punch counts
        self.freq = defaultdict(int)

        # initialize parsing state
        self.inside = False  # whether inside a hole
        self.label = None  # label of hole, when inside a hole
        self.labels = set()  # type: Set[str]  # labels encountered so far
        self.duplicate_labels = set()  # type: Set[str]

        # initialize marker line parsing results
        self.marker = None

    def pre_cell(self, cell: NotebookNode) -> None:
        pass

    def punch_begin_marker_line(self, cell: NotebookNode, line: str) -> None:
        """Process begin marker line."""
        pass

    def punch_end_marker_line(self, cell: NotebookNode, line: str) -> None:
        """Process end marker line."""
        pass

    def punch_inside_line(self, cell: NotebookNode, line: str) -> None:
        """Process line inside hole."""
        pass

    def punch_outside_line(self, cell: NotebookNode, line: str) -> None:
        """Process line outside hole."""
        pass

    def handle_error(self, message: str, cell: NotebookNode, ignore: bool = True) -> None:
        """Handle error with marker.

        :param message: error message
        :param cell: notebook cell with error
        :param ignore: whether to ignore the marker line
        """
        if self.args.allow_errors:
            self.freq['errors'] += 1
            print(message)
            if self.inside:
                self.punch_end_marker_line(cell, "# <<< WARNING: AUTO-GENERATED END MARKER >>>")
                self.inside = False
                self.label = None
                if ignore:
                    self.marker = None
                    action = "  Assuming correct end marker"
                else:
                    action = "  Inserting correct end marker"
            else:
                action = "  Ignoring marker"
                self.marker = None
            print(action)
        else:
            raise AssertionError(message)

    def parse_line(self, line: str, cell_type: str) -> None:
        """Parse given line for given cell type, and decide whether it is a marker line.
        Set instance variable ``marker`` to dictionary with transition, label, and description.
        Sets it to ``None`` if the line is not a marker line.
        Sets ``marker['transition']`` to ``None`` if invalid marker line.

        .. note:: **Modifies**: ``self.marker``

        :param line: marker line
        :param cell_type: type of cell containing marker line
        """
        self.marker = None  # anticipated
        if not re.search(self.args.marker_regex[cell_type], line):
            return

        # recognized as marker line, parse it
        # TODO maybe have separate begin/end regex?
        m = re.search(self.args.marker_line_parsing_regex[cell_type], line)
        # the ``m`` stands for 'match object'
        if m:
            if self.args.label_regex and not re.search(self.args.label_regex, m.group('label')):
                return
            self.marker = {
                'transition': m.group('transition'),
                'label': m.group('label'),
                'description': m.group('description'),
            }
            if TEST:
                print('transition: {transition}, label: {label}, description: {description}'
                      .format(**self.marker))
        else:  # invalid marker line
            self.marker = {'transition': None}

    def punch_line(self, cell: NotebookNode, cell_index: int, line: str, line_index: int) -> None:
        """Punch line.
        """
        # if line is a marker line, process it and update state
        # processing it, yields whether it is BEGIN or END, its label, and description
        # label is needed when filling from source
        if TEST:
            print(' {:>3}| {}'.format(line_index + 1, line))

        self.parse_line(line, cell.cell_type)

        if self.marker:  # it is a marker line
            expected_transitions = sorted(self.args.marker_transitions.values())
            expected_transition = self.args.marker_transitions['end' if self.inside else 'begin']

            # check marker line validity
            if not self.marker['transition']:  # parse failed
                self.handle_error('Invalid marker line in cell {} on line {}:\n  {}'
                                  .format(cell_index, line_index + 1, line),
                                  cell)
            elif self.marker['transition'] not in expected_transitions:
                self.handle_error('Invalid transition in marker line of cell {} on line {}:\n'
                                  '  {}\n  Expected {}, but found {transition}.'
                                  .format(cell_index, line_index + 1,
                                          line, expected_transitions, **self.marker),
                                  cell)
            elif self.marker['transition'] != expected_transition:
                self.handle_error('Unexpected transition in marker line of cell {} on line {}:\n'
                                  '  {}\n  Expected {}{}, but found {transition}.'
                                  .format(cell_index, line_index + 1,
                                          line, expected_transition,
                                          ' [{}]'.format(self.label) if self.inside else '', **self.marker),
                                  cell, False)
            elif self.inside and self.marker['label'] != self.label:
                self.handle_error('Unexpected label in end marker line of cell {} on line {}:\n'
                                  '  {}\n  Expected [{}], but found [{label}]'
                                  .format(cell_index, line_index + 1, line, self.label, **self.marker),
                                  cell)

        if self.marker:
            # update state
            self.inside = not self.inside

            if self.inside:  # begin of hole
                self.freq['holes'] += 1
                self.label = self.marker['label']

                if self.args.list:
                    print('  [{label}] {description}'.format(**self.marker))

                # check uniqueness of label
                if self.label in self.labels:
                    # label already appeared earlier
                    # TODO depending on purpose of hole collection: ignore, warn, error, or abort
                    # TODO maybe put this in a hook method?
                    # TODO maybe use warnings module?
                    if self.args.debug:
                        print('Duplicate label in cell {} on line {}:\n  {}'
                              .format(cell_index, line_index + 1, line))
                    self.duplicate_labels.add(self.label)

                self.labels.add(self.label)

                self.punch_begin_marker_line(cell, line)  # hook method

            else:  # end of hole
                self.punch_end_marker_line(cell, line)  # hook method

            if not self.inside:
                self.label = None

        else:  # not marker line, then it depends on the side
            if self.inside:
                self.punch_inside_line(cell, line)  # hook method
            else:
                self.punch_outside_line(cell, line)  # hook method

    def post_cell(self, cell: NotebookNode) -> None:
        pass

    def post_nb(self) -> Any:
        pass

    def punch_nb(self) -> Any:
        """Template method; main entry point.
        """
        # __init__() took care of initialization

        if self.args.list:
            print('Listing labels and descriptions for "{}"'.format(self.name))

        # traverse cells of notebook
        for cell_index, cell in enumerate(self.nb.cells):
            if TEST:
                ct = cell.cell_type
                print('===== cell {}: {} ====='.format(cell_index, ct))
                print('marker regex: {}'.format(self.args.marker_regex[ct]))

            self.pre_cell(cell)  # hook method

            # process all lines in cell.source
            for line_index, line in enumerate(cell_lines(cell)):
                self.punch_line(cell, cell_index, line, line_index)

            self.post_cell(cell)  # hook method

        # check for missing end marker
        if self.inside:
            message = 'Missing end marker line for last hole labeled [{}].'.format(self.label)
            if self.args.allow_errors:
                self.freq['errors'] += 1
                print(message)
                print("  Assuming correct end marker.")
                # TODO call hook method that finishes filling in case of PunchedNotebookProcessor
                # Without that, the filling may be incomplete
            else:
                raise AssertionError(message)

        if self.duplicate_labels:
            if self.args.verbose:
                print('Labels with duplicates: {}'.format(sorted(self.duplicate_labels)))
            self.freq['labels with duplicates'] = len(self.duplicate_labels)

        return self.post_nb()  # hook method


class PunchedNotebookProcessor(PunchBaseProcessor):
    """Create punched notebook.

    self.nb_punched = result notebook under construction
    self.cell_punched = next cell under construction
    self.source_chad = chad being inserted in current hole
    """

    def __init__(self, nb: NotebookNode, args: Namespace, name: str):
        super().__init__(nb, args, name)
        # initialize result notebook
        self.nb_punched = copy.deepcopy(nb)
        self.nb_punched.cells = []  # to be appended to

        self.source_chad = []  # type: List[NotebookNode]  # chad being inserted to fill current hole
        self.cell_punched = None  # type: Optional[NotebookNode]  # next punched cell being collected

    def pre_cell(self, cell: NotebookNode):
        # create corresponding cell for punched notebook
        # we preserve everything, except possibly the source lines (this could result in empty cells)
        self.cell_punched = copy.deepcopy(cell)
        self.cell_punched.source = []  # List[str] to be converted to single string in post_cell()

    def punch_begin_marker_line(self, cell: NotebookNode, line: str):
        # copy marker line to result notebook, if desired
        if self.args.keep_marker_lines:
            self.cell_punched.source.append(line)

        # take care of filling, if desired
        if self.args.fill:
            # inject filling
            if self.args.punch_source:
                # use chads from source notebook, available in self.args.source_chads
                if self.label in self.args.source_chads:  # corresponding source chad is present
                    # copy the source chad,
                    self.source_chad = self.args.source_chads[self.label][:]  # copy  TODO deep?
                    # it will be injected possibly in multiple (up to three) steps
                    # (1) here, (2) in post_cell(), (3) in punch_end_marker_line()
                    if self.source_chad[0].cell_type == cell.cell_type:
                        # if equal begin cell types, then insert source from first cell here
                        # handle first cell of source chad here (N.B. source chad has first cell)
                        if TEST:
                            print('Copying from begin chad cell to begin hole cell: {}'
                                  .format(repr(self.source_chad[0].source)))
                        self.cell_punched.source.append(self.source_chad[0].source)
                        del self.source_chad[0]  # remove parts of source chad that have been handled
                    else:  # begin cell types differ
                        # insert below this cell (handled in post_cell() and/or punch_end_marker_line())
                        pass
                else:  # corresponding source chad is not present
                    # warn that replacement is not present
                    # TODO find better way of warning
                    if self.args.debug:
                        print('No source chad for label [{}] (not filling this hole).'
                              .format(self.label))
                    self.source_chad = []
                    self.freq['unfilled holes'] += 1

            else:
                # use fixed content, substituting label and description (only useful with source option)
                self.cell_punched.source.append(self.args.filling[cell.cell_type].format(**self.marker))

    def punch_end_marker_line(self, cell: NotebookNode, line: str):
        # check if source chad still has material; if so, handle it; could concern multiple cells
        if self.source_chad:
            # need to split if still in same cell as begin marker line
            if self.cell_punched.source:  # already collected some material preceding this hole
                # append that as cell
                self.cell_punched.source = '\n'.join(self.cell_punched.source)
                self.nb_punched.cells.append(self.cell_punched)
                # reset
                self.cell_punched = copy.deepcopy(cell)
                self.cell_punched.source = []  # to be converted to single string in post_cell()

            # prepend material in source chad, except last if last has same cell type as this cell
            if TEST:
                print('Copying all-but-last cell from chad cell into hole: {} cells'
                      .format(len(self.source_chad[:-1])))
            self.nb_punched.cells.extend(self.source_chad[:-1])
            del self.source_chad[:-1]

            # check if receiver and source end cell have same type
            if self.args.source_chads[self.label][-1].cell_type == cell.cell_type:
                # copy any remaining content
                if TEST:
                    print('Copying from end chad cell to end hole cell: {}'
                          .format(repr(self.source_chad[-1].source)))
                self.cell_punched.source.append(self.source_chad[-1].source)
            else:  # receiving and source end cell types differ; put in a separate cell preceding this one
                if TEST:
                    print('Copying last cell from chad cell into hole: {} cells'
                          .format(len(self.source_chad[:-1])))
                self.nb_punched.cells.append(self.source_chad[-1])
            del self.source_chad[-1]  # now it is empty

        # copy marker line to result notebook, if desired
        if self.args.keep_marker_lines:
            self.cell_punched.source.append(line)

    def punch_outside_line(self, cell: NotebookNode, line: str):
        # copy line to result notebook
        self.cell_punched.source.append(line)

    def post_cell(self, cell: NotebookNode):
        # join source list into single string
        self.cell_punched.source = '\n'.join(self.cell_punched.source)

        if self.cell_punched.source or not self.inside or not self.args.punch_source:
            # transfer non-empty collected cell to result notebook
            self.nb_punched.cells.append(self.cell_punched)
            self.cell_punched = None
        # TODO not satisfactory when not copying marker lines; empty begin cell is suppressed, empty end cell not

    def post_nb(self):
        unused_chads = sorted(set(self.args.source_chads.keys()).difference(self.labels))
        if unused_chads:
            self.freq['unused source chads'] = len(unused_chads)
            if self.args.verbose:
                print('Labels of unused source chads: {}'.format(unused_chads))
        unfilled_labels = sorted(self.labels.difference(self.args.source_chads.keys()))
        if self.args.punch_source and unfilled_labels:
            if self.args.verbose:
                print('Labels of unfilled holes: {}'.format(unfilled_labels))

        return self.freq, self.nb_punched


class ChadsNotebookProcessor(PunchBaseProcessor):
    """Create chads notebook.

    self.nb_chads = result notebook under construction
    self.cell_chads = next cell under construction, with source as list of strings (to be compacted)
    """

    def __init__(self, nb: NotebookNode, args: Namespace, name: str):
        super().__init__(nb, args, name)
        # initialize result notebook
        self.nb_chads = copy.deepcopy(self.nb)
        self.nb_chads.cells = []  # it needs its own list object (avoid aliasing)
        self.cell_chads = None  # type: Optional[NotebookNode]

    def pre_cell(self, cell: NotebookNode):
        # create corresponding cell for chads notebook
        # we preserve everything, except possibly the source lines (this could result in empty cells)
        self.cell_chads = copy.deepcopy(cell)
        self.cell_chads.source = []  # to be converted to single string in post_cell()

    def punch_begin_marker_line(self, cell: NotebookNode, line: str):
        # copy marker line to result notebook, if desired
        if self.args.keep_marker_lines:
            self.cell_chads.source.append(line)

    def punch_end_marker_line(self, cell: NotebookNode, line: str):
        # copy marker line to result notebook, if desired
        if self.args.keep_marker_lines:
            self.cell_chads.source.append(line)

    def punch_inside_line(self, cell: NotebookNode, line: str):
        # copy line to result notebook
        self.cell_chads.source.append(line)

    def post_cell(self, cell: NotebookNode):
        # join source list into single string
        self.cell_chads.source = '\n'.join(self.cell_chads.source)

        # transfer collected cell to result notebook
        self.nb_chads.cells.append(self.cell_chads)

    def post_nb(self):
        return self.freq, self.nb_chads


class SourceChadsProcessor(PunchBaseProcessor):
    """Create chads for later use as source.
    Result is delivered in ``args.source_chads``,
    which is a dictionary that maps hole labels to hole content as list of cells.

    self.source_chads = dictionary that maps label to list of cells
    self.cell_source_chads = next cell under construction, with source as list of strings (to be compacted)
    """

    def __init__(self, nb: NotebookNode, args: Namespace, name: str):
        super().__init__(nb, args, name)
        # initialize punched chads (which could serve as source chads in another punch)
        # self.args.source_chads = {}  # dictionary of label-chad pairs  # commented out to allow multiple source files
        self.cell_source_chads = None

    def pre_cell(self, cell: NotebookNode):
        # start collecting hole content for args.source_chads
        if self.inside:
            self.cell_source_chads = copy.deepcopy(cell)
            self.cell_source_chads.source = []  # list of strings, to be consolidated into single string

    def punch_begin_marker_line(self, cell: NotebookNode, line: str):
        # start collecting hole content for args.source_chads
        self.cell_source_chads = copy.deepcopy(cell)
        self.cell_source_chads.source = []  # list of strings, to be consolidated into single string
        # we don't copy marker lines into args.source_chads
        self.args.source_chads[self.label] = []  # overwrites previous chad, if label not unique

    def punch_end_marker_line(self, cell: NotebookNode, line: str):
        # finish collecting hole content for args.source_chads
        # we don't copy marker lines into args.source_chads
        # join source list into single string
        self.cell_source_chads.source = '\n'.join(self.cell_source_chads.source)

        # transfer collected cell to result
        self.args.source_chads[self.label].append(self.cell_source_chads)
        self.cell_source_chads = None

    def punch_inside_line(self, cell: NotebookNode, line: str):
        # copy line to result
        self.cell_source_chads.source.append(line)

    def post_cell(self, cell: NotebookNode):
        if self.inside:
            # finish collecting hole content for args.source_chads
            # join source list into single string
            self.cell_source_chads.source = '\n'.join(self.cell_source_chads.source)

            # transfer collected cell to result
            self.args.source_chads[self.label].append(self.cell_source_chads)

    def post_nb(self):
        return self.freq



"""
Functions for running

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

import sys
from argparse import Namespace

import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors.execute import CellExecutionError
from nbformat import NotebookNode

from .cleaning import clean_code_output, clean_code_metadata, truncate_output_streams

TEST = False


def ipc_kernel_manager_factory(ipc_path):
    try:
        from jupyter_client import AsyncKernelManager as _BaseKernelManager
    except ImportError:
        from jupyter_client import KernelManager as _BaseKernelManager

    class IPCKernelManager(_BaseKernelManager):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, transport='ipc', ip=ipc_path, **kwargs)

    return IPCKernelManager


def run_nb(nb: NotebookNode, args: Namespace) -> None:
    """Run notebook.

    .. note:: **Modifies**: ``nb``

    :param nb: notebook to run
    :param args: arguments (options)
    """
    # clean up before execution
    if args.clean_before:
        clean_code_output(nb)
        clean_code_metadata(nb, args.clean_before_metadata)

    if args.append_cell:
        nb.cells.append(nbformat.v4.new_code_cell(args.appended_cell))

    ep_kwargs = {
        'timeout': args.timeout,
        'allow_errors': args.allow_errors,
        'interrupt_on_timeout': args.interrupt_on_timeout,
        'record_timing': args.record_timing,
    }

    if args.ipc:
        ep_kwargs['kernel_manager_class'] = ipc_kernel_manager_factory(args.ipc)
    if args.kernel_name:
        ep_kwargs['kernel_name'] = args.kernel_name

    # run notebook
    ep = ExecutePreprocessor(**ep_kwargs)
    try:
        resources = {'metadata': {'path': args.run_path}}  # set working directory
        _ = ep.preprocess(nb, resources)  # nb is executed in-place, locally
    except (CellExecutionError, TimeoutError) as e:  # only possible if not args.allow_errors or if timeout
        if getattr(args, 'assert'):  # args.assert gives syntax error
            raise
        else:
            print('{}: {}'.format(type(e).__name__, e), file=sys.stderr)
    finally:
        # clean up after execution
        if args.clean_after:
            clean_code_metadata(nb, args.clean_after_metadata)

        if args.streams_head >= 0:
            truncate_output_streams(nb, args)



"""
Main for nbstats

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.

Summarize Jupyter notebooks on the command line.
"""

from argparse import _ArgumentGroup, SUPPRESS
from pathlib import Path
from textwrap import dedent
from typing import Optional, List

from nbformat import NotebookNode

from nbtoolbelt.arguments import NegatableAction
from nbtoolbelt.counting import nb_metadata, nb_other_metadata, nb_extra_fields, nb_cell_stats, extract_aggregate
from nbtoolbelt.printing import print_dict
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.toolbaseapp import Tool

TOOL = 'stats'


class StatsTool(Tool):
    """Collect and show a statistical summary of each notebook.
    Also compute some statistics over all notebooks.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            description=dedent("""\
                Summarize Jupyter notebooks on stdout with statistics.
                By default summarizes all notebook data.
                Limit to specific items by passing options.
                """),
            action='summarize'
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Collect and present statistics about notebook nb.

        :param nb: notebook to count
        :param nb_path: path to ``nb``
        :return: sequence of resulting (notebook, notebook-path) pairs
        """
        args = self._args

        output = {}  # to accumulate JSON output

        # always print summary of required notebook metadata
        metadata = nb_metadata(nb)
        print_dict(metadata, 'Notebook metadata')
        output['notebook_metadata'] = metadata

        # print notebook-level optional metadata fields, if desired
        if args.metadata:
            other_metadata = nb_other_metadata(nb)
            print_dict(other_metadata, 'Other notebook metadata fields')
            output['notebook_other_metadata'] = other_metadata

        # print the notebook extra metadata, if desired
        if args.extra:
            extra_fields = nb_extra_fields(nb)
            print_dict(extra_fields, 'EXTRA NOTEBOOK FIELDS')
            output['notebook_extra_fields'] = extra_fields

        cell_stats = nb_cell_stats(nb, args)

        # print desired cell statistics
        keys = []
        if args.cell_types:
            print_dict(cell_stats['cell_types'], "Cell types")
            keys.append('cell_types')
        if args.sources:
            print_dict(cell_stats['sources'], "Cell sources")
            keys.append('sources')
        if args.metadata or args.tags:
            print_dict(cell_stats['cell_metadata'], "Cell metadata fields")
            keys.append('cell_metadata')
        if args.attachments:
            print_dict(cell_stats['cell_attachments'], "Cell attachments")
            keys.append('cell_attachments')
        if args.outputs or args.streams or args.errors:
            print_dict(cell_stats['code_outputs'], "Code cell outputs")
            keys.append('code_outputs')
        if args.execution:
            print_dict(cell_stats['code_execution'], 'Code cell execution')
            keys.append('code_execution')
        if args.extra:
            print_dict(cell_stats['cell_extra'], "EXTRA CELL FIELDS")
            keys.append('cell_extra')
        output.update({key: value for key, value in cell_stats.items() if key in keys})

        self._aggregate['outputs'][-1].update(output)

        return []

    def process_collected_data(self):
        super().process_collected_data()
        # print some aggregation results
        args = self._args

        if len(args.notebooks) <= 1:
            return

        # print header
        print('\nTotals\n======\n')

        aggregate = self._aggregate['outputs'][0]

        # print notebook-level optional metadata fields, if desired
        if args.metadata:
            other_metadata = extract_aggregate(aggregate.get('notebook_other_metadata', {}))
            print_dict(other_metadata, 'Other notebook metadata fields')

        # print the notebook extra metadata, if desired
        if args.extra:
            extra_fields = extract_aggregate(aggregate.get('notebook_extra_fields', {}))
            print_dict(extra_fields, 'EXTRA NOTEBOOK FIELDS')

        cell_stats = aggregate

        # print desired cell statistics
        if args.cell_types:
            cell_types = extract_aggregate(cell_stats.get('cell_types', {}))
            print_dict(cell_types, "Cell types")
        if args.sources:
            cell_sources = extract_aggregate(cell_stats.get('sources', {}))
            print_dict(cell_sources, "Cell sources")
        if args.metadata or args.tags:
            cell_metadata = extract_aggregate(cell_stats.get('cell_metadata', {}))
            print_dict(cell_metadata, "Cell metadata fields")
        if args.attachments:
            cell_attachments = extract_aggregate(cell_stats.get('cell_attachments', {}))
            print_dict(cell_attachments, "Cell attachments")
        if args.outputs or args.streams or args.errors:
            code_outputs = extract_aggregate(cell_stats.get('code_outputs', {}))
            print_dict(code_outputs, "Code cell outputs")
        if args.execution:
            # delete 'maximum In[#]' from collected outputs, because its statistics don't make much sense
            # TODO improve the following deletion; also: it mentions three attributes twice
            if 'code_execution' in cell_stats and 'maximum In[#]' in cell_stats['code_execution']:
                del cell_stats['code_execution']['maximum In[#]']
            code_execution = extract_aggregate(cell_stats.get('code_execution', {}))
            print_dict(code_execution, 'Code cell execution')
        if args.extra:
            cell_extra = extract_aggregate(cell_stats.get('cell_extra', {}))
            print_dict(cell_extra, "EXTRA CELL FIELDS")

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('--all',
                           action=NegatableAction, dest='all_stats', default=SUPPRESS,
                           help='show all statistics' +
                           ' (default: {})'.format(self._args.all_stats)),
        group.add_argument('-c', '--cell-types',
                           action=NegatableAction, default=SUPPRESS,
                           help='count cell types' +
                           ' (default: {})'.format(self._args.cell_types))
        group.add_argument('-s', '--sources',
                           action=NegatableAction, default=SUPPRESS,
                           help='statistics for cell sources' +
                           ' (default: {})'.format(self._args.sources))
        group.add_argument('-m', '--metadata',
                           action=NegatableAction, default=SUPPRESS,
                           help='show notebook metadata and count cell metadata' +
                           ' (default: {})'.format(self._args.metadata))
        group.add_argument('-t', '--tags',
                           action=NegatableAction, default=SUPPRESS,
                           help='count individual cell tags' +
                           ' (default: {})'.format(self._args.tags))
        group.add_argument('-a', '--attachments',
                           action=NegatableAction, default=SUPPRESS,
                           help='count cell attachment MIME types' +
                           ' (default: {})'.format(self._args.attachments))
        group.add_argument('-o', '--outputs',
                           action=NegatableAction, default=SUPPRESS,
                           help='count code cell outputs' +
                           ' (default: {})'.format(self._args.outputs))
        group.add_argument('--streams',
                           action=NegatableAction, default=SUPPRESS,
                           help='count code cell output stream names' +
                           ' (default: {})'.format(self._args.streams))
        group.add_argument('-e', '--errors',
                           action=NegatableAction, default=SUPPRESS,
                           help='count code cell error names' +
                           ' (default: {})'.format(self._args.errors))
        group.add_argument('-x', '--execution',
                           action=NegatableAction, default=SUPPRESS,
                           help='statistics for code execution' +
                           ' (default: {})'.format(self._args.execution))
        group.add_argument('--extra',
                           action=NegatableAction, default=SUPPRESS,
                           help='report extra fields outside metadata' +
                           ' (default: {})'.format(self._args.extra))

    def check_and_adjust_arguments(self):
        args = self._args

        # if --all, summarize everything
        if args.all_stats:
            args.cell_types = True
            args.sources = True
            args.metadata = True
            args.tags = True
            args.attachments = True
            args.outputs = True
            args.streams = True
            args.errors = True
            args.execution = True
            args.extra = True


def main(cli_args: Optional[List[str]] = None) -> int:
    return StatsTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Version, easily accessible in other files

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

# TODO: should each tool have its own version?

import importlib.metadata

#: Package name
package_name = 'nbtoolbelt'

#: Version number, as string
try:
    __version__ = importlib.metadata.version(package_name)
except importlib.metadata.PackageNotFoundError:
    __version__ = 'unknown'



"""
Functions for counting elements in Jupyter notebooks

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from argparse import Namespace
from collections import defaultdict
from typing import Any, Tuple, Dict, Mapping, Callable

import numpy as np
from nbformat import NotebookNode

# TODO these should also be easily available elsewhere, e.g. in cleaning
# TODO consider which can be turned into a frozenset
# TODO consider putting these in the embedded config file
REQUIRED_NB_FIELDS = {"metadata", "nbformat_minor", "nbformat", "cells"}
REQUIRED_NB_METADATA_FIELDS = {"kernelspec", "language_info"}
CELL_TYPES = ('markdown', 'code', 'raw',)
REQUIRED_CELL_FIELDS = {
    'markdown': {"cell_type", "metadata", "source"},
    'code': {"cell_type", "metadata", "source", "execution_count", "outputs"},
    'raw': {"cell_type", "metadata", "source"}
}
OPTIONAL_CELL_FIELDS = {
    'markdown': {"attachments"},
    'code': set(),
    'raw': {"attachments"}
}
OPTIONAL_OUTPUT_TYPES = {
    'execute_result': {'data', 'metadata', 'execution_count'},
    'stream': {'name', 'text'},
    'display_data': {'data', 'metadata', },
    'error': {'ename', 'evalue', 'traceback'},
}

# TODO replace 'sources' by 'source', because that is the field name in the JSON file
#: Fields used in the dictionary returned by ``nb_cell_stats()``.
CELL_STATISTICS = (
    'cell_types',  #: cell type counts
    'sources',  #: cell sources counts
    'cell_metadata',  #: cell metadata counts, including separate ``tags``
    'cell_attachments',  #: cell attachment MIME type counts, and total
    'code_execution',  #: code cell execution count statistics
    'code_outputs',  #: code cell counts per output_type, subcounts per ``stream`` and ``error``, and total
    'cell_extra',  #: counts for extra (unknown) fields in cells
)

# dictionary keys for source statistics
EMPTY_SOURCES = 'total empty sources'
SOURCE_LINES = 'total source lines'
SOURCE_WORDS = 'total source words'
SOURCE_CHARS = 'total source chars'
EMPTY_SOURCES_MD = 'markdown empty sources'
SOURCE_LINES_MD = 'markdown source lines'
SOURCE_WORDS_MD = 'markdown source words'
SOURCE_CHARS_MD = 'markdown source chars'
EMPTY_SOURCES_CODE = 'code empty sources'
SOURCE_LINES_CODE = 'code source lines'
SOURCE_WORDS_CODE = 'code source words'
SOURCE_CHARS_CODE = 'code source chars'
EMPTY_SOURCES_RAW = 'raw empty sources'
SOURCE_LINES_RAW = 'raw source lines'
SOURCE_WORDS_RAW = 'raw source words'
SOURCE_CHARS_RAW = 'raw source chars'

# dictionary keys for code cell outputs statistics
EMPTY_OUTPUTS = 'code cells without outputs'
TOTAL_OUTPUT_COUNT = 'total output count'

# dictionary keys for code_execution cell statistics
EMPTY_EXEC_COUNT = 'not executed'
MAX_EXEC_COUNT = 'maximum In[#]'
LAST_EXEC_COUNT_IN_ORDER = 'executed in linear order'
EXECUTED = 'executed'
NOT_EXEC_IN_ORDER = 'not executed in linear order'


def count_source(source: str) -> Tuple[int, int, int]:
    """Count number of non-blank lines, words, and non-whitespace characters.

    :param source: string to count
    :return: number of non-blank lines, words, and non-whitespace characters
    """
    lines = [line for line in source.split('\n') if line and not line.isspace()]
    words = source.split()
    chars = ''.join(words)

    return len(lines), len(words), len(chars)


def nb_metadata(nb: NotebookNode) -> Dict[str, Any]:
    """Summarize notebook global metadata.

    :param nb: notebook to inspect
    :return: dictionary with format, kernel, and language info
    """
    result = {'format version': '{:1d}.{:1d}'.format(nb.get('nbformat', 1), nb.get('nbformat_minor', 0))}
    if 'kernelspec' in nb.metadata:
        result['kernel'] = nb.metadata.kernelspec.name
    if 'language_info' in nb.metadata:
        l_i = nb.metadata.language_info
        result['language'] = l_i.name + ' ' + l_i.version
    return result


def nb_extra_fields(nb: NotebookNode) -> Dict[str, int]:
    """Extract extra global fields in notebook.

    :param nb: notebook to inspect
    :return: list of extra fields (not set, because that is not JSON serializable)
    """
    return {key: 1 for key in nb.keys() if key not in REQUIRED_NB_FIELDS}


def nb_other_metadata(nb: NotebookNode) -> Dict[str, int]:
    """Extract other global metadata fields in notebook.

    :param nb: notebook to inspect
    :return: set of global metadata fields other than required
    """
    return {key: 1 for key in nb.metadata.keys() if key not in REQUIRED_NB_METADATA_FIELDS}


def nb_cell_stats(nb: NotebookNode, args: Namespace = None) -> Dict[str, Dict[str, int]]:
    """Count occurrences of various elements in notebook cells.

    If ``args`` is not ``None``, then the following boolean arguments are used
    (if present; absent is interpreted as ``False``):

    * sources
    * metadata
    * tags
    * code
    * streams
    * errors

    If ``args`` is ``None``, then all statistics are gathered.

    :param nb: notebook to inspect
    :param args: namespace with arguments; if None, count everything
    :return: dictionary of dictionaries with counts per section;
        each section has its own key; see CELL_STATISTICS
    """
    # process the notebook cells
    result = {key: defaultdict(int) for key in CELL_STATISTICS}
    all_executed_in_order = True  # whether all code cells so far were executed in order

    # traverse all cells and gather statistics
    for index, cell in enumerate(nb.cells):  # index can be used for debug output
        result['cell_types']['total cell count'] += 1  # count all cells
        ct = cell.cell_type
        result['cell_types'][ct] += 1  # count each cell type

        # compute source statistics
        empty_cell = True  # in case of missing source (should not happen)
        if getattr(args, 'sources', False):
            lines, words, chars = count_source(cell.source)  # cell.source should always be present
            empty_cell = chars == 0
            if empty_cell:
                result['sources'][EMPTY_SOURCES] += 1
                if ct == 'markdown':
                    result['sources'][EMPTY_SOURCES_MD] += 1
                elif ct == 'code':
                    result['sources'][EMPTY_SOURCES_CODE] += 1
                elif ct == 'raw':
                    result['sources'][EMPTY_SOURCES_RAW] += 1
            if chars:
                result['sources'][SOURCE_LINES] += lines
                result['sources'][SOURCE_WORDS] += words
                result['sources'][SOURCE_CHARS] += chars
                if ct == 'markdown':
                    result['sources'][SOURCE_LINES_MD] += lines
                    result['sources'][SOURCE_WORDS_MD] += words
                    result['sources'][SOURCE_CHARS_MD] += chars
                elif ct == 'code':
                    result['sources'][SOURCE_LINES_CODE] += lines
                    result['sources'][SOURCE_WORDS_CODE] += words
                    result['sources'][SOURCE_CHARS_CODE] += chars
                elif ct == 'raw':
                    result['sources'][SOURCE_LINES_RAW] += lines
                    result['sources'][SOURCE_WORDS_RAW] += words
                    result['sources'][SOURCE_CHARS_RAW] += chars

        # count each metadata key
        for attr in cell.metadata:  # cell.metadata should always be present
            if args is None or getattr(args, 'metadata', False) or (getattr(args, 'tags', False) and attr == 'tags'):
                result['cell_metadata'][attr] += 1

        # count each tag in tags metadata
        if (args is None or getattr(args, 'tags', False)) and 'tags' in cell.metadata:
            for tag in cell.metadata.tags:
                result['cell_metadata']['tag ' + tag] += 1

        # count each attachment mime type
        if 'attachments' in cell and getattr(args, 'attachments', False):
            result['cell_attachments']['total count of cells with attachments'] += 1
            for attachment in cell.attachments.values():
                for key in attachment:
                    result['cell_attachments']['total attachments count'] += 1
                    result['cell_attachments'][key] += 1

        if ct == 'code':  # process code cell
            ec = cell.get('execution_count')
            if type(ec) is int:
                result['code_execution'][EXECUTED] += 1
                # update maximum execution count
                if ec > result['code_execution'][MAX_EXEC_COUNT]:
                    result['code_execution'][MAX_EXEC_COUNT] = ec
                # update whether all executed in order
                if all_executed_in_order and ec == result['code_execution'][LAST_EXEC_COUNT_IN_ORDER] + 1:
                    result['code_execution'][LAST_EXEC_COUNT_IN_ORDER] = ec
                elif empty_cell:
                    pass  # ignore; should not happen; but nbconvert --execute can produce this
                else:
                    all_executed_in_order = False
                    result['code_execution'][NOT_EXEC_IN_ORDER] += 1
            else:  # it should be None
                result['code_execution'][EMPTY_EXEC_COUNT] += 1
                if not empty_cell:
                    all_executed_in_order = False
                    result['code_execution'][NOT_EXEC_IN_ORDER] += 1

            # process outputs of code cell (should be empty if ec is None)
            if not cell.outputs:  # empty code output
                result['code_outputs'][EMPTY_OUTPUTS] += 1
            for output in cell.outputs:
                outputs = getattr(args, 'outputs', False)
                if args is None or outputs:
                    result['code_outputs'][TOTAL_OUTPUT_COUNT] += 1
                ot = output['output_type']
                streams = getattr(args, 'streams', False)
                errors = getattr(args, 'errors', False)
                if args is None or outputs or (streams and ot == 'stream') or (errors and ot == 'error'):
                    result['code_outputs'][ot] += 1
                if (args is None or streams) and ot == 'stream':
                    result['code_outputs'][ot + ' ' + output['name']] += 1
                if (args is None or errors) and ot == 'error':
                    result['code_outputs'][ot + ' ' + output.ename] += 1

        # count non-standard fields in cells
        for field in cell:
            if field not in REQUIRED_CELL_FIELDS[ct].union(OPTIONAL_CELL_FIELDS[ct]):
                result['cell_extra'][field] += 1

    # result['code_execution'][LAST_EXEC_COUNT_IN_ORDER] is a special case
    # it can be set to 0 by inspection in if-condition above, and remain on 0
    if result['code_execution'][LAST_EXEC_COUNT_IN_ORDER] == 0:
        del result['code_execution'][LAST_EXEC_COUNT_IN_ORDER]

    return result


def nb_code_execution_stats(nb: NotebookNode) -> Dict:
    """Count number of (executed) code cells and errors in notebook.

    :param nb: notebook to inspect
    :return: dictionary with
        'code cells': number of code cells,
        'executed': number of executed cells,
        'with errors': number of execution errors
    """
    result = {  # counts
        'code cells': 0,
        'code cells with empty source': 0,
        'executed': 0,
        'with errors': 0,
    }

    # TODO: introduce constants for the keys in result?
    for cell in nb.cells:
        if cell.cell_type == 'code':
            result['code cells'] += 1
            # N.B. attribute execution_count only, and always, exists in code cells
            result['code cells with empty source'] += int(count_source(cell.source)[2] == 0)
            result['executed'] += int(cell.execution_count is not None)
            # error occurred when cell.outputs contains an output containing output_type 'error'
            result['with errors'] += int(any(output.output_type == 'error' for output in cell.outputs))

    return result


def extract_aggregate(d: Dict[str, Dict[str, Any]], attr: str = 'total',
                      convert: Callable[[Any], Any] = int) -> Dict[str, Any]:
    """Extract converted values for given attribute from nested dictionary.
    """
    return {key: convert(value[attr]) for key, value in d.items()}


PAIRING_TOKEN = ' : '  #: character sequence to combine key pairs


def pair_keys(key1: str, key2: str) -> str:
    """Pair two strings into one string,
    where key2 may also have been paired.

    :param key1: first key
    :param key2: second key
    :return: paired key
    """
    return key1 + PAIRING_TOKEN + key2


def is_key_pair(key: str) -> bool:
    """Return whether key is paired,
    that is, result of ``pair_key``.

    :param key: key to test
    :return: whether key is result of ``pair_key()``
    """
    return ' : ' in key


def unpair_key(key: str) -> Tuple[str, str]:
    """Unpair key into key1, key2 such that
    key1 is not a pair, and ``pair_keys(key1, key2) == key``.

    .. note:: **Assumption**: ``is_key_pair(key)``

    :param key: key to split
    :return: key1, key2 where ``pair_keys(key1, key2) == key``
    """
    i = key.find(PAIRING_TOKEN)
    if i < 0:
        raise ValueError('Pairing token ("{}") missing'.format(PAIRING_TOKEN))
    else:
        return key[:i], key[i + len(PAIRING_TOKEN):]


def flatten_mapping(mapping: Mapping[str, Any]) -> Dict[str, Any]:
    """Return recursively flattened mapping as dict.

    :param mapping: object to flatten
    :return: flattened mapping
    """
    if hasattr(mapping, 'get'):
        result = {}
        for key, value in mapping.items():
            if hasattr(value, 'get'):
                result.update({pair_keys(key, key2): value2 for key2, value2 in value.items()})
            else:
                result.update({key: value})
        return result
    else:
        raise ValueError('Mapping has no "get" attribute')  # return mapping


def unflatten_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    """Return recursively unflattened dictionary,
    that is, ``flatten_mapping(result) == d``.

    :param d: dictionary to unflatten
    :return: unflattened version of d
    """
    # all key-value pairs in d, where key == pair_keys(key1, key2)
    # with the same key1, give rise to one pair key1: { key2: value, ...}
    result = defaultdict(dict)

    for key, value in d.items():
        if is_key_pair(key):
            key1, key2 = unpair_key(key)
            result[key1].update({key2: value})
        else:
            result.update({key: value})

    # now, recursively unflatten the resulting values
    return {key: (unflatten_dict(value) if type(d) is defaultdict else value) for key, value in result.items()}


def clean_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    """Remove key-nan pairs and convert non-nan np types to native Python.

    :param d: dict to clean
    :return: cleaned dict
    """
    result = {}

    for key, value in d.items():
        if isinstance(value, dict):
            kv = {key: clean_dict(value)}
        elif isinstance(value, np.integer):  # TODO: check warning for np.integer
            kv = {key: int(value)}
        elif isinstance(value, float):
            if np.isnan(value):
                kv = {}
            else:
                kv = {key: float(value)}
        else:
            kv = {key: value}
        result.update(kv)

    return result



"""
Base class for concrete tools

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

import json
import sys
from argparse import Namespace, ArgumentParser, _ArgumentGroup, SUPPRESS
from pathlib import Path
from typing import Any, Optional, List, Dict

import pandas as pd
from nbformat import NotebookNode

from .arguments import LoadConfigFile, ExtendAction, NegatableAction, TestAction
from .config import load_config
from .counting import flatten_mapping, unflatten_dict, clean_dict
from .notebook_io import read_nb, write_nb
from .printing import print_namespace
from .processing import ProcessingResultType
from .running import run_nb
from .validating import validate_nb

TOOL = 'base'

TEST = False


class Tool(object):
    """Base class for concrete tools.
    It is in itself a functional tool that copies notebooks.

    These methods are template methods:

    * ``main()``
    * ``parse_args()``
    * ``process_file()``

    Concrete tools inherit from Tool and override these (hook) methods:

    * ``__init__()`` MUST ``super()`` at begin, when overriding
    * ``config_tool_args_parsing()`` optional; to configure parser for tool-specific arguments
    * ``check_and_adjust_arguments()`` optional
    * ``print_tool_args()`` optional; to print tool-specific arguments (called only in verbose mode)
    * ``process_nb()`` compulsory; to process one notebook; MUST NOT CALL ``super()``
    * ``process_collected_data()`` optional; MUST CALL ``super()`` at begin, when overriding

    About the instance variables:
    """

    OUTPUT_HEADER = ":" * 14  # to set off file names in output

    def __init__(self,
                 name: str = 'nb' + TOOL,
                 action: str = 'copy',
                 description: str = 'Copy Jupyter notebooks.'):
        """Initialize the tool base.

        Overriding method must call ``super()`` at begin.

        :param name: the tool's name (e.g., used to find parameters in configuration files)
        :param action: verb (infinitive) describing tool's main action (usable after 'to') as help
        :param description: short text describing the tool (for usage instruction)
        """
        self._name = name
        self._description = description
        self._action = action
        # load default options from embedded configuration file
        self._args = load_config(self._name)  # type: Namespace  # TODO when to pass in verbose=True?
        # anticipated exit code; tools can set this
        self._exit_code = 0
        # dictionary with results aggregated over multiple notebooks
        # concrete tools initialize their part in __init__()
        self._aggregate = {
            'outputs': [{'file_count': 0}],  # type: List[Dict[str, Any]]
            # outputs[0] is overall; outputs[i] for i > 0 is per notebook argument
            # outputs[0].file_count == number of notebook files fully processed so far
        }

    def tool_name(self) -> str:
        """Get the tool's name.

        :return: the tool's name
        """
        return self._name

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Process notebook nb.

        Can update ``self._aggregate``.

        Hook method to be overridden in concrete tools.
        Overriding method must NOT call ``super()``.


        .. note:: **Modifies**: ``nb``, ``self._aggregate``

        :param nb: notebook to process
        :param nb_path: path to ``nb``
        :return: sequence of resulting (notebook, notebook-path) pairs
        """
        # Copy notebook
        print('Copying: {}'.format(nb_path.name))

        # set up destination path
        if self._args.inplace:
            nb_result_path = nb_path
        else:
            nb_result_path = nb_path.with_name(nb_path.stem + '-copy' + nb_path.suffix)

        return [(nb, nb_result_path)]

    def process_file(self, nb_path: Path) -> None:
        """Process one file, given by its path.

        .. note:: **Modifies**: ``self._aggregate``

        :param nb_path: path of file to process
        :return: exit code
        """
        # read the notebook
        nb = read_nb(nb_path, self._args)
        if nb is None:
            self._exit_code = 1
            return

        # process the notebook
        try:
            # validate the notebook (if requested and it makes sense)
            if self._args.validate and self._name != 'nbvalidate':
                valid = validate_nb(nb, getattr(self._args, 'assert'))
                if not valid:
                    print('Notebook is INVALID')

            # run the notebook (if requested and it makes sense)
            if self._args.run and self._name not in {'nbvalidate', 'nbrun', 'nbhead'}:
                run_nb(nb, self._args)

            processing_results = self.process_nb(nb, nb_path)  # type: ProcessingResultType
        except Exception as e:
            e_name = type(e).__name__
            print('Processing of "{}" failed ({}):\n  {}'.format(nb_path.name, e_name, e), file=sys.stderr)
            if self._args.debug:
                print('  While processing notebook from file:', nb_path.resolve(), file=sys.stderr)
                raise
            else:
                self._exit_code = 1
                return

        # write all resulting notebook files that were returned in processing_results
        if processing_results:
            if self._args.write_files:
                written = set()

                for res_nb, res_nb_path in processing_results:
                    if write_nb(res_nb, res_nb_path, self._args) is None:
                        self._exit_code = 1
                        return
                    written.update({res_nb_path.name})

                if self._args.verbose:
                    print('  Files written: {}'.format(written))
            else:
                print('No file{} written'.format('' if len(processing_results) == 1 else 's'))

    def process_collected_data(self) -> None:
        """Process outputs collected from all processed notebooks.

        Hook method to be overridden in concrete tools.
        Overriding method must call ``super()`` at begin.
        """
        # accumulate int values for all files
        data = [flatten_mapping(d) for d in self._aggregate['outputs'][1:]]
        if data:
            df = pd.DataFrame(data)
            statistics = df.describe(include='all')
            # append row with total for each numeric column
            sums = df.sum(numeric_only=True)
            sums.name = 'total'
            statistics = pd.concat([statistics, sums.to_frame().T])
            # update the aggregate entry of outputs (item 0)
            self._aggregate['outputs'][0].update(clean_dict(unflatten_dict(statistics.to_dict())))

    def write_output(self) -> None:
        """Write accumulated output to JSON.
        """
        fp = Path(self._args.output_json)
        if self._args.debug:
            print('Writing JSON output to:', fp)
        try:
            with fp.open('w', encoding='utf-8') as json_file:
                json.dump(self._aggregate['outputs'], json_file, indent=2, sort_keys=True)
        except Exception as e:
            print('Could not write JSON output for {} to: {}'.format(self._name, fp), file=sys.stderr)
            print('  {}: {}'.format(type(e).__name__, e), file=sys.stderr)
            # TODO should this exit with exit code > 0?
            if self._args.debug:
                raise

    def parser_with_common_arguments(self) -> ArgumentParser:
        """Create argument parser and add common arguments to it.

        :return: the pre-configured parser
        """
        parser = ArgumentParser(
            prog='nbtb ' + self._name[2:],
            description=self._description,
            add_help=True,
        )

        parser.add_argument('notebooks', metavar='NB.ipynb', type=str, nargs='*',
                            action=ExtendAction, default=SUPPRESS,
                            help='notebooks to ' + self._action)

        # TODO reconsider meaning of verbose and quiet together
        parser.add_argument('-v', '--verbose',
                            action=NegatableAction, default=SUPPRESS,
                            help='verbose mode produces extra output' +
                                 ' (default: {})'.format(self._args.verbose))
        parser.add_argument('-q', '--quiet',
                            action=NegatableAction, default=SUPPRESS,
                            help='quiet mode produces less output' +
                                 ' (default: False)'.format(self._args.quiet))
        parser.add_argument('--assert', action=NegatableAction,
                            help='assert mode: when processing fails, abort with exit code 1' +
                                 ' (default: {})'.format(getattr(self._args, 'assert')))
        parser.add_argument('--validate', action=NegatableAction,
                            help='validate notebook before processing' +
                                 ' (default: {})'.format(self._args.validate))
        parser.add_argument('--run', action=NegatableAction,
                            help='run notebook before processing' +
                                 ' (default: {})'.format(self._args.run))
        parser.add_argument('--inplace',
                            action=NegatableAction, dest='inplace', default=SUPPRESS,
                            help='replace original notebooks with processing result' +
                                 ' (default: {})'.format(self._args.inplace))
        parser.add_argument('--write-files', action=NegatableAction,
                            help='do write result files' +
                                 ' (default: {})'.format(self._args.write_files))
        parser.add_argument('-n', action='store_false', dest='write_files', default=SUPPRESS,
                            help='short for --no-write-files: '
                                 'do processing but do not write result files (dry run)')
        parser.add_argument('--config',
                            action=LoadConfigFile, type=Path, const=self._name, default=SUPPRESS,
                            metavar='FILE.json',
                            help='read configuration from FILE.json (in JSON)')
        parser.add_argument('--output-json',
                            action='store', metavar='FILE.json',
                            help='write statistical output to FILE.json')
        parser.add_argument('-d', '--debug',
                            action='store_true',
                            help='debug mode produces diagnostic output (default: False)')
        if TEST:
            parser.add_argument('--test',
                                action=TestAction, nargs='?', const='my_const', default=0,
                                choices=['a', 'b'], required=True, metavar='TEST', dest='my_dest',
                                help="this option prints the action arguments")

        return parser

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        """Add argument configurations that are tool specific to the given argument group.

        Hook method to be overridden in concrete tools.

        :param group: argument group to add into
        """
        pass

    def check_and_adjust_arguments(self):
        """Do tool-specific checks and adjustments of parsed arguments.

        Hook method to be overridden in concrete tools.
        """
        pass

    def print_tool_args(self) -> None:
        """Print tool-specific arguments; used especially in verbose mode.
        Indent the lines by 2 spaces.

        Hook method to be overridden in concrete tools.
        """
        pass

    def parse_args(self, arguments: List[str]) -> None:
        """Configure an argument parser and parse the command-line arguments,
        updating the namespace with the parsed arguments.

        .. note:: **Modifies**: ``self._args``

        :param arguments: list of arguments on command line
        """
        # set up argument parser
        parser = self.parser_with_common_arguments()

        # add tool specific arguments
        group = parser.add_argument_group('optional arguments specific for tool \'{}\''.format(self._name[2:]))
        self.config_tool_args_parsing(group)

        # parse arguments
        self._args = parser.parse_args(arguments, self._args)

        # do tool-specific checking and adjusting of parsed arguments
        self.check_and_adjust_arguments()

        if self._args.debug:
            print_namespace(self._args, 'All option settings')

        if self._args.inplace:
            print('Replacing original notebooks')
        if not self._args.write_files:
            print('Dry run (no files written)')

        if self._args.verbose:
            print('Options for {}:'.format(self._name))
            self.print_tool_args()

    def main(self, cli_args: Optional[List[str]] = None) -> int:
        """Main entry point.
        Processes all files in ``args.notebooks``, returning an exit code (0 = success).

        Can set ``self._exit_code`` to signal abortion.

        :param cli_args: list of command-line arguments (None when used as independent script)
        """
        if cli_args is None:
            # independent script; pick up command-line arguments, dropping program name
            cli_args = sys.argv[1:]

        # parse arguments into namespace self._args
        try:
            self.parse_args(cli_args)
            # TODO: transfer options (incl. e.g. freq) to JSON output: self._aggregate['outputs'][0]
            # TODO: could be tool specific!
        except Exception as e:
            print('Argument parsing failed: {}'.format(e), file=sys.stderr)
            if getattr(getattr(self, '_args'), 'debug', False):
                raise e
            else:
                self._exit_code = 1
                return self._exit_code

        # check if all files exist
        for fn in self._args.notebooks:
            if not Path(fn).exists():
                print("File not found: {}".format(fn), file=sys.stderr)
                self._exit_code = 1
                return self._exit_code

        # process all notebook arguments
        for fn in self._args.notebooks:
            nb_path = Path(fn)
            self._aggregate['outputs'].append({"file_name": nb_path.name})  # TODO: check typing

            # print header if more than one file, or in verbose mode
            if (len(self._args.notebooks) > 1 or self._args.verbose) and not self._args.quiet:
                # print file names with colons (imitating nbshow from nbdime)
                print(self.OUTPUT_HEADER)
                print(nb_path.resolve() if self._args.debug else nb_path.name)
                print(self.OUTPUT_HEADER)

            # process the file
            self.process_file(nb_path)

            if self._exit_code > 0:
                break

            self._aggregate['outputs'][0]['file_count'] += 1

        self.process_collected_data()

        if self._args.output_json:
            self.write_output()

        file_count = self._aggregate['outputs'][0]['file_count']
        if self._args.verbose:
            print("\nNotebooks processed: {}".format(file_count))

        return self._exit_code


def main(cli_args: Optional[List[str]] = None) -> int:
    return Tool().main(cli_args)


if __name__ == "__main__":
    sys.exit(Tool().main())



"""
Tool nbhead; shows head of notebooks

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.

This tool also illustrates how to extend class Tool, to define a new concrete tool.
"""

from argparse import _ArgumentGroup, SUPPRESS
from pathlib import Path
from typing import Optional, List

from nbformat import NotebookNode

from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.toolbaseapp import Tool

TOOL = 'head'


class HeadTool(Tool):
    """Show the first *n* (default 5) source lines of the first cell of each notebook.
    If *n* is negative, show the last *n* source lines of the last cell.

    It overrides:

    * ``config_tool_args_parsing()`` to configure parser for tool-specific arguments
    * ``print_tool_args()`` to print tool-specific arguments (used in verbose mode)
    * ``process_nb()`` to process one notebook
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            description='Show head or tail of Jupyter notebooks.',
            action=TOOL
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Show head of notebook.

        :param nb: notebook to show head of
        :param nb_path: notebook's path
        :return: empty result
        """
        if nb.cells:
            source = ''
            if self._args.number >= 0:
                if nb.cells[0].source:
                    source = nb.cells[0].source.split('\n')[:self._args.number]  # list of first n source lines
            else:
                if nb.cells[-1].source:
                    source = nb.cells[-1].source.split('\n')[self._args.number:]  # list of last n source lines
            if source:
                print('\n'.join(source))

        return []  # no notebooks produced

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-#', '--number', type=int, default=SUPPRESS,
                           help='number of head/tail lines to show of first/last cell; < 0 for tail' +
                                ' (default: {})'.format(self._args.number))

    def print_tool_args(self) -> None:
        print('  Show {0} {1} source lines of {0} cell'.format(
            'first' if self._args.number >= 0 else 'last',
            abs(self._args.number)
        ))


def main(cli_args: Optional[List[str]] = None) -> int:
    return HeadTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Tool nbview

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

import os
import webbrowser
from argparse import _ArgumentGroup, SUPPRESS
from pathlib import Path
from typing import Optional, List

from nbformat import NotebookNode

from nbtoolbelt.arguments import NegatableAction
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.rendering import render_nb
from nbtoolbelt.toolbaseapp import Tool

TOOL = 'view'


class ViewTool(Tool):
    """Show notebooks in the browser, with embedded MathJax and attachments rendered.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            description='View Jupyter notebooks your browser.',
            action=TOOL
        )
        self._aggregate['written_html'] = []  # list of opened html files

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """View notebook.
        That is, convert it to html, and open it in the default browser.

        .. note:: **Side effect**: Creates html file

        :param nb: notebook to show head of
        :param nb_path: notebook's path
        :return: empty
        """
        # determine destination path of html file
        name_insert = '' if self._args.inplace else self._args.view_result_name
        html_path = nb_path.with_name(nb_path.stem + name_insert + '.html')

        # convert to html
        html = render_nb(nb, self._args)

        if self._args.write_files:
            # write html to file
            try:
                with html_path.open('w', encoding='utf-8') as fp:
                    fp.write(html)
            except Exception as e:
                e_name = type(e).__name__
                print('Writing of "{}" failed ({}): {}.'.format(html_path.name, e_name, e), file=sys.stderr)
                if self._args.debug:
                    print('  Tried to write html file:', html_path.resolve(), file=sys.stderr)
                raise

            # register in self._aggregate, for deletion (if so desired)
            if self._args.wait_delete:
                self._aggregate['written_html'].append(html_path)
        else:
            print('No file written')  # TODO is this really necessary?

        # open in browser
        if self._args.browser:
            webbrowser.open('file://' + Path(html_path).resolve().as_posix(), new=2)

        return []  # no notebooks produced

    def process_collected_data(self):
        if self._aggregate['written_html']:
            answer = input('Delete created html files ([y]/n)? ')
            deleting = not answer or answer.lower()[0] == 'y'
            for file in self._aggregate['written_html']:
                if deleting:
                    if not self._args.quiet:
                        print('Deleting:', file)
                    os.remove(file)
                else:
                    if not self._args.quiet:
                        print('Keeping:', file)

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('--template-file', default=SUPPRESS,
                           metavar='FILE',
                           help='[DEPRECATED] template file')
        group.add_argument('-t', '--template', default=SUPPRESS,
                           metavar='NAME',
                           help='template to use' +
                                ' (default: {})'.format(self._args.template))
        group.add_argument('-b', '--browser', action=NegatableAction, default=SUPPRESS,
                           help='open html in browser' +
                                ' (default: {})'.format(self._args.browser))
        group.add_argument('-w', '--wait-delete', action=NegatableAction, default=SUPPRESS,
                           help='interactively wait for user confirmation to delete html files' +
                                ' (default: {})'.format(self._args.wait_delete))

    def print_tool_args(self) -> None:
        print('  Template: {!r}'.format(self._args.template))
        print('  Open in web browser: {!r}'.format(self._args.browser))
        print('  Wait for user to confirm deletion of html: {!r}'.format(self._args.wait_delete))

    def check_and_adjust_arguments(self):
        args = self._args
        if not args.write_files:
            args.browser = False
            args.wait_delete = False
        if args.template_file:
            import warnings
            warnings.warn(f"The template-file option is deprecated; value {self._args.template_file!r} is ignored",
                          UserWarning)


def main(cli_args: Optional[List[str]] = None) -> int:
    return ViewTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
NbConvert Preprocessor to inline attachments

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.

Parts of this code are reworked from the rendernb checklet for Momotor <momotor.org>.
"""

import base64
from typing import Any, Dict, Tuple
from urllib.parse import quote

from nbconvert.preprocessors import Preprocessor
from nbformat import NotebookNode

TEST = False


class InlineAttachmentsPreprocessor(Preprocessor):
    """To embed cell attachments inside the cell's source.
    """

    ACCEPTABLE_TYPES = (
        'image/png',
        'image/jpeg',
        'image/gif',
        'image/webp',
        'image/bmp',
    )

    def preprocess_cell(self, cell: NotebookNode, resources: Dict[str, Any], index: int) -> \
            Tuple[NotebookNode, Dict[str, Any]]:
        """Take a Jupyter Notebook cell and inline all attachments as data.

        .. note:: **Modifies**: ``cell``

        :param cell: cell for which to inline attachments
        :param resources: global resources
        :param index: cell index in notebook
        :return: modified cell, updated resources
        """
        if 'attachments' in cell:
            remaining_attachments = {}

            for name, data in cell.attachments.items():
                data_type, data_content = None, None
                # data_content = next((data[data_type] for data_type in data if data_type in ACCEPTABLE_TYPES), None)
                for data_type in data:
                    if data_type in self.ACCEPTABLE_TYPES:
                        if TEST:
                            print("inline_attachments acceptable data type: {}".format(data_type))
                        data_content = data[data_type]
                        break

                if data_content:
                    # Re-encode base64 to strip whitespace
                    try:
                        decoded_content = base64.b64decode(data_content)
                    except TypeError:
                        if TEST:
                            print("inline_attachment decoding gave TypeError")
                        continue

                    recoded_content = base64.b64encode(decoded_content)
                    if TEST:
                        print("inline_attachments recoded content")

                    new_uri = 'data:{};base64,{}'.format(quote(data_type), quote(recoded_content))
                    if TEST:
                        print("inline_attachments new URI: {}".format(new_uri[:100]))
                    for att_name in (name, quote(name)):
                        old_uri = 'attachment:' + att_name
                        if TEST:
                            print("inline_attachments old URI: {}".format(old_uri))
                        cell.source = cell.source.replace(old_uri, new_uri)

                    if TEST:
                        print("inline_attachments new cell source:\n{}".format(cell.source[:1000]))

                else:
                    remaining_attachments[name] = data

            cell['attachments'] = remaining_attachments

        return cell, resources



"""
Common processing definitions

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from pathlib import Path
from typing import Sequence, Tuple

from nbformat import NotebookNode

# type alias for processing result
ProcessingResultType = Sequence[Tuple[NotebookNode, Path]]



"""
Tool nbpunch

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from argparse import _ArgumentGroup, SUPPRESS
from pathlib import Path
from typing import Optional, List

from nbformat import NotebookNode

from nbtoolbelt.arguments import split_csv, join_csv, NegatableAction, LoadSourceFile
from nbtoolbelt.counting import CELL_TYPES
from nbtoolbelt.printing import print_dict
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.punching import punch_via_tags, PunchedNotebookProcessor, ChadsNotebookProcessor
from nbtoolbelt.toolbaseapp import Tool

TOOL = 'punch'


class PunchTool(Tool):
    """Punch notebooks, optionally filling holes with material coming from other notebooks,
    writing the results to new notebooks.
    """

    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            description="""Punch holes in Jupyter notebooks.""",
            action=TOOL
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Punch notebook nb.

        .. note:: **Modifies**: ``nb``

        :param nb: notebook to punch
        :param nb_path: path to ``nb``
        :return: sequence of resulting (notebook, notebook-path) pairs
        """
        args = self._args
        freq = {}

        if args.tags:
            freq, nb_punched, nb_chads = punch_via_tags(nb, args)

        # set up destination paths and result
        result = []

        if args.punched:
            if not args.tags:
                freq_punched, nb_punched = PunchedNotebookProcessor(nb, args, nb_path.name).punch_nb()
                freq.update(freq_punched)
            nb_punched_path = nb_path.with_name(nb_path.stem + args.punch_punched_result_name + nb_path.suffix)
            result.append((nb_punched, nb_punched_path))

        if args.chads:
            nb_chads_path = nb_path.with_name(nb_path.stem + args.punch_chads_result_name + nb_path.suffix)
            if not args.tags:
                freq_chads, nb_chads = ChadsNotebookProcessor(nb, args, nb_chads_path.name).punch_nb()
                freq.update(freq_chads)
            result.append((nb_chads, nb_chads_path))

        # print statistics
        if not args.quiet:
            print_dict(freq, 'Punch cell statistics')
        self._aggregate['outputs'][-1].update(freq)

        return result

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        args = self._args

        group.add_argument('-t', '--tags',
                           dest='tags', type=split_csv, default=SUPPRESS,
                           metavar='TAGS',
                           help='comma-separated list of tags that trigger removal of source from cells;' +
                                " e.g. YourTurn (default: '{}')".format(join_csv(args.tags)))
        group.add_argument('-p', '--punched',
                           action=NegatableAction, default=SUPPRESS,
                           help='whether to write punched notebook' +
                                ' (default: {})'.format(args.punched))
        group.add_argument('-c', '--chads',
                           action=NegatableAction, default=SUPPRESS,
                           help='whether to write chads notebook' +
                                ' (default: {})'.format(args.chads))
        group.add_argument('-m', '--keep-marker-lines',
                           action=NegatableAction, default=SUPPRESS,
                           help='whether to keep marker lines' +
                                ' (default: {})'.format(args.keep_marker_lines))
        group.add_argument('-f', '--fill',
                           action=NegatableAction, default=SUPPRESS,
                           help='whether to fill punched holes' +
                                ' (default: {})'.format(args.fill))
        group.add_argument('-s', '--source',
                           action=LoadSourceFile, dest='punch_source', default=SUPPRESS,
                           help='source notebook for filling holes; implies --fill' +
                                ' (default: {!r})'.format(args.punch_source))
        group.add_argument('-l', '--list',
                           action=NegatableAction, default=SUPPRESS,
                           help='list marker labels and descriptions' +
                                ' (default: {})'.format(args.list))
        group.add_argument('-e', '--allow-errors',
                           action=NegatableAction, default=SUPPRESS,
                           help='continue on parsing errors in marker lines' +
                                ' (default: {})'.format(args.allow_errors))
        group.add_argument('--label-regex',
                           dest='label_regex', default=SUPPRESS,
                           help='only process marker lines whose labels match REGEX' +
                                ' (default: {!r}; empty regex matches all)'.format(args.label_regex))

    def check_and_adjust_arguments(self):
        args = self._args

        if args.punch_source:
            # --source implies --fill
            args.fill = True

        if args.tags and args.punch_source:
            raise AssertionError('Cannot use --source with --tags.')

    def print_tool_args(self) -> None:
        args = self._args

        if args.tags:
            print('  Tags that trigger punch: {}'.format(args.tags))
        else:
            if args.debug:
                print('  Markers that trigger punch:')
                for cell_type in CELL_TYPES:
                    print('    {} regex: {}'.format(cell_type, repr(args.marker_regex[cell_type])))
                print('  Marker line parsing to extract transition, label, and description:')
                for cell_type in CELL_TYPES:
                    print('    {} regex: {}'.format(cell_type, repr(args.marker_line_parsing_regex[cell_type])))
                print('  Marker transitions:')
                for transition in 'begin', 'end':
                    print('    {}: {}'.format(transition, repr(args.marker_transitions[transition])))
        print('  Writing punched notebook: {}'.format(args.punched))
        print('  Writing chads notebook: {}'.format(args.chads))
        print('  Keeping marker lines: {}'.format(args.keep_marker_lines))
        print('  Listing labels and descriptions: {}'.format(args.list))
        if args.fill:
            if args.punch_source and not args.tags:
                print('  Fill holes from source notebook: {}'.format(args.punch_source))
            else:  # no source, or using tags
                print('  Fill holes as follows:')
                for cell_type in CELL_TYPES:
                    print('    {}: {}'.format(cell_type, repr(args.filling[cell_type])))
        else:
            print('  Do not fill punched holes')
        if args.allow_errors:
            action = 'Not stopping at errors'
        else:
            action = 'Stopping at first error'
        print('  {}'.format(action))


def main(cli_args: Optional[List[str]] = None) -> int:
    return PunchTool().main(cli_args)


if __name__ == "__main__":
    import sys

    sys.exit(main())



"""
Main for nbrun

Copyright (c) 2017 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""

from argparse import _ArgumentGroup, SUPPRESS
from pathlib import Path
from typing import Optional, List

from nbformat import NotebookNode

from nbtoolbelt.arguments import NegatableAction, quote_help
from nbtoolbelt.counting import nb_code_execution_stats
from nbtoolbelt.printing import print_dict
from nbtoolbelt.processing import ProcessingResultType
from nbtoolbelt.running import run_nb
from nbtoolbelt.toolbaseapp import Tool

TOOL = 'run'


class RunTool(Tool):
    """Run each notebook, with optional pre-/post-cleaning, and return results.
    """
    def __init__(self) -> None:
        super().__init__(
            name='nb' + TOOL,
            action=TOOL,
            description="""Run Jupyter notebooks, with optional pre/post-cleaning."""
        )

    def process_nb(self, nb: NotebookNode, nb_path: Path) -> ProcessingResultType:
        """Run notebook nb.

        .. note:: **Modifies**: ``nb``

        :param nb: notebook to run
        :param nb_path: path to ``nb``
        :return: resulting (notebook, notebook-path) pair
        """
        args = self._args

        run_nb(nb, args)

        # print statistics, unless in quiet mode
        stats = nb_code_execution_stats(nb)
        if not args.quiet:
            print_dict(stats, "Execution statistics")
        self._aggregate['outputs'][-1].update(stats)

        # set up destination path
        if args.inplace:
            nb_run_path = nb_path
        else:
            nb_run_path = nb_path.with_name(nb_path.stem + args.run_result_name + nb_path.suffix)

        if args.debug:  # TODO is this needed here (or should it happen in Tool.process_file?
            print('Destination of run:', nb_run_path)

        return [(nb, nb_run_path)]

    def config_tool_args_parsing(self, group: _ArgumentGroup) -> None:
        group.add_argument('-e', '--allow-errors',
                           action=NegatableAction,
                           help='continue on errors' +
                                ' (default: {})'.format(self._args.allow_errors))
        group.add_argument('-b', '--clean-before',
                           action=NegatableAction,
                           help='clean code output before running' +
                                ' (default: {}'.format(self._args.clean_before))
        group.add_argument('-a', '--clean-after',
                           action=NegatableAction,
                           help='clean code metadata after running' +
                                ' (default: {})'.format(self._args.clean_after))
        group.add_argument('-k', '--kernel-name',
                           dest='kernel_name', default=SUPPRESS,
                           help='name of kernel to use, e.g. python3; '' for kernel specified in notebook;' +
                                " default: '{}')".format(self._args.kernel_name))
        group.add_argument('-p', '--run-path',
                           dest='run_path', default=SUPPRESS,
                           help='working directory for execution ;"" for current working directory' +
                                " (default: '{}')".format(self._args.run_path))
        group.add_argument('-t', '--timeout',
                           type=int, default=SUPPRESS,
                           help='timeout time in seconds per cell; -1 for unlimited' +
                                ' (default: {})'.format(self._args.timeout))
        group.add_argument('-i', '--interrupt-on-timeout',
                           action=NegatableAction, default=SUPPRESS,
                           help='interrupt execution on timeout' +
                                ' (default: {})'.format(self._args.interrupt_on_timeout))
        group.add_argument('-r', '--record-timing',
                           action=NegatableAction, default=SUPPRESS,
                           help='record execution times' +
                                ' (default: {})'.format(self._args.record_timing))
        group.add_argument('-w', '--append-cell',
                           action=NegatableAction, default=SUPPRESS,
                           help='append special code cell' +
                                " (default: '{}') ".format(quote_help(self._args.appended_cell)) +
                                'before executing notebooks' +
                                ' (default: {})'.format(self._args.append_cell))
        group.add_argument('--streams-head',
                           type=int, default=SUPPRESS,
                           help='limit output streams to a maximum number of lines per cell; -1 for unlimited' +
                                ' (default: {})'.format(self._args.streams_head))
        group.add_argument('--ipc',
                           default=SUPPRESS,
                           help='Use IPC with given path to communicate with the kernel; \'\' for TCP' +
                                " (default: '{}')".format(self._args.ipc))

    def print_tool_args(self) -> None:
        if self._args.clean_before:
            action = 'Cleaning'
        else:
            action = 'Keeping'
        what = f" and {self._args.clean_before_metadata} metadata" if self._args.clean_before_metadata else ''
        print('  {} code output{} before running'.format(action, what))

        print('  Timeout:', 'unlimited' if self._args.timeout < 0 else '{} s'.format(self._args.timeout))
        print('  Kernel name:', self._args.kernel_name if self._args.kernel_name else 'from notebook')
        print('  Working dir:', self._args.run_path if self._args.run_path else 'current directory')

        if self._args.allow_errors:
            action = 'Not stopping at errors'
        else:
            action = 'Stopping at first error'
        print('  {}'.format(action))

        if self._args.timeout >= 0:
            action = 'times out after {} seconds'.format(self._args.timeout)
        else:
            action = 'does not time out'
        print('  Cell execution {}'.format(action))

        if self._args.timeout >= 0:
            if self._args.interrupt_on_timeout:
                action = 'I'
            else:
                action = 'Not i'
            print('  {}nterrupting kernel on timeout'.format(action))

        if self._args.record_timing:
            action = 'R'
        else:
            action = 'Not r'
        print('  {}ecording execution times'.format(action))

        if self._args.append_cell:
            action = 'Appending'
        else:
            action = 'Not appending'
        print('  {} code cell with %whos'.format(action))

        if self._args.clean_after:
            action = 'Cleaning {}'.format(self._args.clean_after_metadata)
        else:
            action = 'Keeping'
        print('  {} code metadata after running'.format(action))

        if self._args.streams_head >= 0:
            print('  Limiting output streams to {} lines'.format(self._args.streams_head))
        else:
            print('  Not limiting output streams')


def main(cli_args: Optional[List[str]] = None) -> int:
    return RunTool().main(cli_args)


if __name__ == "__main__":
    import sys
    sys.exit(main())



"""
Exported interface, when nbtoolbelt is imported as library

Copyright (c) 2017-2020 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.

.. data:: __version__

   Version number, as string

.. data:: version_info

   Version as tuple

.. data:: package_name

   Name of the ``nbtoolbelt`` package

.. data:: CELL_STATISTICS

   Tuple with the keys used in the dictionary returned by ``nb_cell_stats()``
"""

from ._version import __version__, package_name
from .counting import CELL_STATISTICS, nb_metadata, nb_other_metadata, nb_extra_fields, nb_cell_stats
from .printing import print_dict, print_set
from .running import run_nb

__all__ = [
    "__version__", "package_name",
    "CELL_STATISTICS",
    "nb_metadata", "nb_other_metadata", "nb_extra_fields", "nb_cell_stats",  # from counting
    "print_dict", "print_set",  # from printing
    "run_nb",  # from running
    ]
